@@ -7,11 +7,10 @@ use piston::event_loop::{EventSettings, Events};
 use piston::input::{RenderEvent, UpdateEvent};
 use piston::window::WindowSettings;
 use piston::{EventLoop, IdleEvent};
-use std::env;
 use std::net::SocketAddr;
+use structopt::StructOpt;
 
 const FPS: u64 = 60;
-const NUM_PLAYERS: usize = 2;
 const INPUT_SIZE: usize = std::mem::size_of::<u8>();
 
 const WINDOW_HEIGHT: u32 = 800;
@@ -19,17 +18,31 @@ const WINDOW_WIDTH: u32 = 600;
 
 mod box_game;
 
-fn main() -> Result<(), Box<dyn std::error::Error>> {
-    // read cmd line arguments very clumsily
-    let args: Vec<String> = env::args().collect();
-    assert_eq!(args.len(), 3);
+#[derive(StructOpt)]
+struct Opt {
+    #[structopt(short, long)]
+    local_port: u16,
+    #[structopt(short, long)]
+    num_players: usize,
+    #[structopt(short, long)]
+    host: SocketAddr,
+}
 
-    let port: u16 = args[1].parse()?;
-    let host_addr: SocketAddr = args[2].parse()?;
+fn main() -> Result<(), Box<dyn std::error::Error>> {
+    // read cmd line arguments
+    let opt = Opt::from_args();
 
     // create a GGRS session for a spectator
-    let mut sess =
-        ggrs::start_p2p_spectator_session(NUM_PLAYERS as u32, INPUT_SIZE, port, host_addr)?;
+    let mut sess = ggrs::start_p2p_spectator_session(
+        opt.num_players as u32,
+        INPUT_SIZE,
+        opt.local_port,
+        opt.host,
+    )?;
+
+    // change catch-up parameters, if desired
+    sess.set_max_frames_behind(5)?; // when the spectator is more than this amount of frames behind, it will catch up
+    sess.set_catchup_speed(2)?; // set this to 1 if you don't want any catch-ups
 
     // start the GGRS session
     sess.start_session()?;
@@ -45,15 +58,8 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
             .build()
             .unwrap();
 
-    // load a font to render text
-    let assets = find_folder::Search::ParentsThenKids(3, 3)
-        .for_folder("assets")
-        .unwrap();
-    let freetype = ft::Library::init().unwrap();
-    let font = assets.join("FiraSans-Regular.ttf");
-
     // Create a new box game
-    let mut game = box_game::BoxGame::new(font);
+    let mut game = box_game::BoxGame::new(opt.num_players);
     let mut gl = GlGraphics::new(opengl);
 
     // event settings
@@ -64,15 +70,14 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
 
     // event loop
     while let Some(e) = events.next(&mut window) {
-        // render
+        // render update
         if let Some(args) = e.render_args() {
-            game.render(&mut gl, &freetype, &args);
+            game.render(&mut gl, &args);
         }
 
-        // game update
+        // game update - tell GGRS it is time to advance the frame and handle the requests
         if let Some(_) = e.update_args() {
             if sess.current_state() == SessionState::Running {
-                // tell GGRS it is time to advance the frame and handle the requests
                 match sess.advance_frame() {
                     Ok(requests) => game.handle_requests(requests),
                     Err(GGRSError::PredictionThreshold) => {
@@ -86,14 +91,14 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
         // idle
         if let Some(_args) = e.idle_args() {
             sess.poll_remote_clients();
-        }
 
-        // handle GGRS events
-        for event in sess.events() {
-            println!("Event: {:?}", event);
-            if let GGRSEvent::Disconnected { .. } = event {
-                println!("Disconnected from host.");
-                return Ok(());
+            // handle GGRS events
+            for event in sess.events() {
+                println!("Event: {:?}", event);
+                if let GGRSEvent::Disconnected { .. } = event {
+                    println!("Disconnected from host.");
+                    return Ok(());
+                }
             }
         }
     }
