@@ -1,35 +1,41 @@
 extern crate freetype as ft;
 
-use ggrs::{PlayerType, SessionState};
 use glutin_window::GlutinWindow as Window;
 use opengl_graphics::{GlGraphics, OpenGL};
 use piston::event_loop::{EventSettings, Events};
 use piston::input::{RenderEvent, UpdateEvent};
 use piston::window::WindowSettings;
 use piston::{Button, EventLoop, Key, PressEvent, ReleaseEvent};
+use structopt::StructOpt;
 
 const FPS: u64 = 60;
-const NUM_PLAYERS: usize = 2;
 const INPUT_SIZE: usize = std::mem::size_of::<u8>();
-const CHECK_DISTANCE: u32 = 7;
 
 const WINDOW_HEIGHT: u32 = 800;
 const WINDOW_WIDTH: u32 = 600;
 
 mod box_game;
 
+#[derive(StructOpt)]
+struct Opt {
+    #[structopt(short, long)]
+    num_players: usize,
+    #[structopt(short, long)]
+    check_distance: u32,
+}
+
 fn main() -> Result<(), Box<dyn std::error::Error>> {
+    // read cmd line arguments
+    let opt = Opt::from_args();
+
     // create a GGRS session with two players
-    let mut sess = ggrs::start_synctest_session(NUM_PLAYERS as u32, INPUT_SIZE, CHECK_DISTANCE)?;
+    let mut sess =
+        ggrs::start_synctest_session(opt.num_players as u32, INPUT_SIZE, opt.check_distance)?;
 
-    // add player - this is a synctest, we skip the second player
-    sess.add_player(PlayerType::Local, 0)?;
-
-    // set input delay for the local player
-    sess.set_frame_delay(2, 0)?;
-
-    // start the GGRS session
-    sess.start_session()?;
+    // set input delay for any player you want
+    for i in 0..opt.num_players {
+        sess.set_frame_delay(2, i)?;
+    }
 
     // Change this to OpenGL::V2_1 if not working
     let opengl = OpenGL::V3_2;
@@ -42,15 +48,8 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
             .build()
             .unwrap();
 
-    // load a font to render text
-    let assets = find_folder::Search::ParentsThenKids(3, 3)
-        .for_folder("assets")
-        .unwrap();
-    let freetype = ft::Library::init().unwrap();
-    let font = assets.join("FiraSans-Regular.ttf");
-
     // Create a new box game
-    let mut game = box_game::BoxGame::new(font);
+    let mut game = box_game::BoxGame::new(opt.num_players);
     let mut gl = GlGraphics::new(opengl);
 
     // event settings
@@ -61,42 +60,49 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
 
     // event loop
     while let Some(e) = events.next(&mut window) {
-        // render
+        // render update
         if let Some(args) = e.render_args() {
-            game.render(&mut gl, &freetype, &args);
+            game.render(&mut gl, &args);
         }
 
         // game update
         if let Some(_) = e.update_args() {
-            // do stuff only when the session is ready
-            if sess.current_state() == SessionState::Running {
-                // tell GGRS it is time to advance the frame and handle the requests
-                let local_input = game.local_input();
-                let requests = sess.advance_frame(0, &local_input)?;
-
-                // handle requests
-                game.handle_requests(requests);
+            // create inputs for all players
+            let mut all_inputs = Vec::new();
+            for i in 0..opt.num_players {
+                all_inputs.push(game.local_input(i));
             }
+            // tell GGRS it is time to advance the frame and handle the requests
+            let requests = sess.advance_frame(&all_inputs)?;
+            game.handle_requests(requests);
         }
 
-        // update key state
+        // key state update
         if let Some(Button::Keyboard(key)) = e.press_args() {
             match key {
                 Key::W => game.key_states[0] = true,
                 Key::A => game.key_states[1] = true,
                 Key::S => game.key_states[2] = true,
                 Key::D => game.key_states[3] = true,
+                Key::Up => game.key_states[4] = true,
+                Key::Left => game.key_states[5] = true,
+                Key::Down => game.key_states[6] = true,
+                Key::Right => game.key_states[7] = true,
                 _ => (),
             }
         }
 
-        // update key state
+        // key state update
         if let Some(Button::Keyboard(key)) = e.release_args() {
             match key {
                 Key::W => game.key_states[0] = false,
                 Key::A => game.key_states[1] = false,
                 Key::S => game.key_states[2] = false,
                 Key::D => game.key_states[3] = false,
+                Key::Up => game.key_states[4] = false,
+                Key::Left => game.key_states[5] = false,
+                Key::Down => game.key_states[6] = false,
+                Key::Right => game.key_states[7] = false,
                 _ => (),
             }
         }
