@@ -6,7 +6,7 @@ use crate::network::udp_msg::{
 };
 use crate::network::udp_socket::NonBlockingSocket;
 use crate::sessions::p2p_session::{
-    Event, DEFAULT_DISCONNECT_NOTIFY_START, DEFAULT_DISCONNECT_TIMEOUT,
+    Event, DEFAULT_DISCONNECT_NOTIFY_START, DEFAULT_DISCONNECT_TIMEOUT, DEFAULT_FPS,
 };
 use crate::time_sync::TimeSync;
 use crate::{Frame, PlayerHandle, NULL_FRAME};
@@ -28,7 +28,6 @@ const SYNC_RETRY_INTERVAL: Duration = Duration::from_millis(200);
 const RUNNING_RETRY_INTERVAL: Duration = Duration::from_millis(200);
 const KEEP_ALIVE_INTERVAL: Duration = Duration::from_millis(200);
 const QUALITY_REPORT_INTERVAL: Duration = Duration::from_millis(200);
-const MAX_SEQ_DISTANCE: u16 = 1 << 15;
 const MAX_PAYLOAD: usize = 467; // 512 is max safe UDP payload, minus 45 bytes for the rest of the packet
 
 fn millis_since_epoch() -> u128 {
@@ -67,6 +66,7 @@ pub(crate) struct UdpProtocol {
     disconnect_timeout: Duration,
     disconnect_notify_start: Duration,
     shutdown_timeout: Instant,
+    fps: u32,
 
     // the other client
     peer_addr: SocketAddr,
@@ -91,8 +91,6 @@ pub(crate) struct UdpProtocol {
     round_trip_time: u128,
     last_send_time: Instant,
     last_recv_time: Instant,
-    send_seq: u16,
-    recv_seq: u16,
 }
 
 impl PartialEq for UdpProtocol {
@@ -143,6 +141,7 @@ impl UdpProtocol {
             disconnect_timeout: DEFAULT_DISCONNECT_TIMEOUT,
             disconnect_notify_start: DEFAULT_DISCONNECT_NOTIFY_START,
             shutdown_timeout: Instant::now(),
+            fps: DEFAULT_FPS,
 
             // the other client
             peer_addr,
@@ -167,8 +166,6 @@ impl UdpProtocol {
             round_trip_time: 0,
             last_send_time: Instant::now(),
             last_recv_time: Instant::now(),
-            send_seq: 0,
-            recv_seq: 0,
         }
     }
 
@@ -185,7 +182,7 @@ impl UdpProtocol {
         }
         // Estimate which frame the other client is on by looking at the last frame they gave us plus some delta for the packet roundtrip time.
         let ping = i32::try_from(self.round_trip_time).expect("Ping is higher than i32::MAX");
-        let remote_frame = self.last_received_input.frame + (ping * 60 / 1000);
+        let remote_frame = self.last_received_input.frame + ((ping * self.fps as i32) / 1000);
         // Our frame "advantage" is how many frames behind the remote client we are. (It's an advantage because they will have to predict more often)
         self.local_frame_advantage = remote_frame - local_frame;
     }
@@ -198,6 +195,11 @@ impl UdpProtocol {
         self.disconnect_notify_start = notify_start;
     }
 
+    pub(crate) fn set_fps(&mut self, fps: u32) {
+        assert!(fps > 0);
+        self.fps = fps;
+    }
+
     pub(crate) fn network_stats(&self) -> Option<NetworkStats> {
         if self.state != ProtocolState::Synchronizing && self.state != ProtocolState::Running {
             return None;
@@ -325,12 +327,6 @@ impl UdpProtocol {
         }
     }
 
-    fn next_sequence_number(&mut self) -> u16 {
-        let ret = self.send_seq;
-        self.send_seq += 1;
-        ret
-    }
-
     /*
      *  SENDING MESSAGES
      */
@@ -430,11 +426,7 @@ impl UdpProtocol {
 
     fn queue_message(&mut self, body: MessageBody) {
         // set the header
-        let header = MessageHeader {
-            magic: self.magic,
-            sequence_number: self.next_sequence_number(),
-        };
-
+        let header = MessageHeader { magic: self.magic };
         let msg = UdpMessage { header, body };
 
         self.packets_sent += 1;
@@ -450,31 +442,17 @@ impl UdpProtocol {
      */
 
     pub(crate) fn handle_message(&mut self, msg: &UdpMessage) {
+        // don't handle messages if shutdown
         if self.state == ProtocolState::Shutdown {
             return;
         }
 
-        // filter messages that don't match what we expect
-        match &msg.body {
-            MessageBody::SyncRequest(_) | MessageBody::SyncReply(_) => {
-                // filter packets that don't match the magic if we have set if yet
-                if self.remote_magic != 0 && msg.header.magic != self.remote_magic {
-                    return;
-                }
-            }
-            _ => {
-                // filter packets that don't match the magic
-                if msg.header.magic != self.remote_magic {
-                    return;
-                }
-                // filter out-of-order packets
-                if msg.header.sequence_number - self.recv_seq > MAX_SEQ_DISTANCE {
-                    return;
-                }
-            }
+        // filter packets that don't match the magic if we have set it already
+        if self.remote_magic != 0 && msg.header.magic != self.remote_magic {
+            return;
         }
-        // update sequence number of received packages
-        self.recv_seq = msg.header.sequence_number;
+
+        // update time when we last received packages
         self.last_recv_time = Instant::now();
 
         // if the connection has been marked as interrupted, send an event to signal we are receiving again
@@ -483,6 +461,7 @@ impl UdpProtocol {
             self.event_queue.push_back(Event::NetworkResumed);
         }
 
+        // handle the message
         match &msg.body {
             MessageBody::SyncRequest(body) => self.on_sync_request(*body),
             MessageBody::SyncReply(body) => self.on_sync_reply(msg.header, *body),
@@ -547,13 +526,12 @@ impl UdpProtocol {
         } else {
             // update the peer connection status
             for i in 0..self.peer_connect_status.len() {
-                assert!(
-                    body.peer_connect_status[i].last_frame
-                        >= self.peer_connect_status[i].last_frame
-                );
                 self.peer_connect_status[i].disconnected = body.peer_connect_status[i].disconnected
                     || self.peer_connect_status[i].disconnected;
-                self.peer_connect_status[i].last_frame = body.peer_connect_status[i].last_frame;
+                self.peer_connect_status[i].last_frame = std::cmp::max(
+                    self.peer_connect_status[i].last_frame,
+                    body.peer_connect_status[i].last_frame,
+                );
             }
         }
 
@@ -570,6 +548,8 @@ impl UdpProtocol {
             return;
         }
 
+        self.running_last_input_recv = Instant::now();
+
         // we know everything is correct, so we decode
         let recv_inputs = decode(&self.last_received_input, body.start_frame, &body.bytes)
             .expect("decoding failed");
@@ -579,10 +559,8 @@ impl UdpProtocol {
             if game_input.frame <= self.last_received_input.frame {
                 continue;
             }
-
             // send the input to the session
             self.last_received_input = *game_input;
-            self.running_last_input_recv = Instant::now();
             self.event_queue.push_back(Event::Input(*game_input));
         }
 
