@@ -5,7 +5,10 @@ use crate::network::udp_msg::ConnectionStatus;
 use crate::network::udp_protocol::UdpProtocol;
 use crate::network::udp_socket::NonBlockingSocket;
 use crate::sync_layer::SyncLayer;
-use crate::{Frame, GGRSEvent, GGRSRequest, PlayerHandle, PlayerType, SessionState, NULL_FRAME};
+use crate::{
+    Frame, GGRSEvent, GGRSRequest, PlayerHandle, PlayerType, SessionState, MAX_PREDICTION_FRAMES,
+    NULL_FRAME,
+};
 
 use std::collections::vec_deque::Drain;
 use std::collections::HashMap;
@@ -16,8 +19,10 @@ use std::time::Duration;
 /// The minimum amounts of frames between sleeps to compensate being ahead of other players
 const RECOMMENDATION_INTERVAL: Frame = 40;
 const MAX_EVENT_QUEUE_SIZE: usize = 100;
+const DEFAULT_SAVE_MODE: bool = false;
 pub(crate) const DEFAULT_DISCONNECT_TIMEOUT: Duration = Duration::from_millis(2000);
 pub(crate) const DEFAULT_DISCONNECT_NOTIFY_START: Duration = Duration::from_millis(500);
+pub(crate) const DEFAULT_FPS: u32 = 60;
 
 #[derive(Debug, PartialEq, Eq)]
 enum Player {
@@ -98,6 +103,10 @@ pub struct P2PSession {
     input_size: usize,
     /// The sync layer handles player input queues and provides predictions.
     sync_layer: SyncLayer,
+    /// FPS defines the expected update frequency of this session.
+    fps: u32,
+    /// With sparse saving, the session will only request to save the minimum confirmed frame.
+    sparse_saving: bool,
 
     /// The time until a remote player gets disconnected.
     disconnect_timeout: Duration,
@@ -145,6 +154,8 @@ impl P2PSession {
             state: SessionState::Initializing,
             num_players,
             input_size,
+            fps: DEFAULT_FPS,
+            sparse_saving: DEFAULT_SAVE_MODE,
             socket,
             local_connect_status,
             next_recommended_sleep: 0,
@@ -174,7 +185,11 @@ impl P2PSession {
     ) -> Result<PlayerHandle, GGRSError> {
         // currently, you can only add players in the init phase
         if self.state != SessionState::Initializing {
-            return Err(GGRSError::InvalidRequest);
+            return Err(GGRSError::InvalidRequest {
+                info:
+                    "Session already started. You can only add players before starting the session."
+                        .to_owned(),
+            });
         }
 
         // add the player depending on type
@@ -191,13 +206,17 @@ impl P2PSession {
     pub fn start_session(&mut self) -> Result<(), GGRSError> {
         // if we are not in the initialization state, we already started the session at some point
         if self.state != SessionState::Initializing {
-            return Err(GGRSError::InvalidRequest);
+            return Err(GGRSError::InvalidRequest {
+                info: "Session already started.".to_owned(),
+            });
         }
 
         // check if all players are added
         for player_handle in 0..self.num_players as PlayerHandle {
             if self.players.get(&player_handle).is_none() {
-                return Err(GGRSError::InvalidRequest);
+                return Err(GGRSError::InvalidRequest{
+                    info: "Not enough players have been added. Keep registering players up to the defined player number.".to_owned(),
+                });
             }
         }
 
@@ -219,7 +238,9 @@ impl P2PSession {
     pub fn disconnect_player(&mut self, player_handle: PlayerHandle) -> Result<(), GGRSError> {
         match self.players.get_mut(&player_handle) {
             // the local player cannot be disconnected
-            None | Some(Player::Local) => Err(GGRSError::InvalidRequest),
+            None | Some(Player::Local) => Err(GGRSError::InvalidRequest {
+                info: "Local Player cannot be disconnected.".to_owned(),
+            }),
             // a remote player can only be disconnected if not already disconnected, since there is some additional logic attached
             Some(Player::Remote(_)) => {
                 if !self.local_connect_status[player_handle].disconnected {
@@ -261,7 +282,7 @@ impl P2PSession {
         // player is not a local player
         match self.players.get(&local_player_handle) {
             Some(Player::Local) => (),
-            _ => return Err(GGRSError::InvalidRequest),
+            _ => return Err(GGRSError::InvalidHandle),
         }
 
         // session is not running and synchronzied
@@ -269,28 +290,56 @@ impl P2PSession {
             return Err(GGRSError::NotSynchronized);
         }
 
+        // This list of requests will be returned to the user
         let mut requests = Vec::new();
 
-        // check game consistency and rollback, if necessary
-        if let Some(first_incorrect) = self
-            .sync_layer
-            .check_simulation_consistency(self.disconnect_frame)
-        {
-            self.adjust_gamestate(first_incorrect, &mut requests);
-            self.disconnect_frame = NULL_FRAME;
+        // if we are in the first frame, we have to save the state
+        if self.sync_layer.current_frame() == 0 {
+            requests.push(self.sync_layer.save_current_state());
         }
 
         // find the total minimum confirmed frame and propagate disconnects
-        let min_confirmed_frame = self.min_confirmed_frame();
+        let min_confirmed = self.min_confirmed_frame();
+
+        // check game consistency and rollback, if necessary.
+        // The disconnect frame indicates if a rollback is necessary due to a previously disconnected player
+        let first_incorrect = self
+            .sync_layer
+            .check_simulation_consistency(self.disconnect_frame);
+        if first_incorrect != NULL_FRAME {
+            self.adjust_gamestate(first_incorrect, min_confirmed, &mut requests);
+            self.disconnect_frame = NULL_FRAME;
+        }
+
+        // in sparse saving mode, we need to make sure not to lose the last saved frame
+        let last_saved = self.sync_layer.last_saved_frame();
+        if self.sparse_saving
+            && self.sync_layer.current_frame() - last_saved >= MAX_PREDICTION_FRAMES as i32
+        {
+            // check if the current frame is confirmed, otherwise we need to roll back
+            if min_confirmed >= self.sync_layer.current_frame() {
+                // the current frame is confirmed, save it
+                requests.push(self.sync_layer.save_current_state());
+            } else {
+                // roll back to the last saved state, resimulate and save on the way
+                self.adjust_gamestate(last_saved, min_confirmed, &mut requests);
+            }
+
+            // after all this, we should have saved the confirmed state
+            assert_eq!(
+                self.sync_layer.last_saved_frame(),
+                std::cmp::min(min_confirmed, self.sync_layer.current_frame())
+            );
+        }
 
         // send confirmed inputs to remotes
-        self.send_confirmed_inputs_to_spectators(min_confirmed_frame);
+        self.send_confirmed_inputs_to_spectators(min_confirmed);
 
         // set the last confirmed frame and discard all saved inputs before that frame
         self.sync_layer
-            .set_last_confirmed_frame(min_confirmed_frame);
+            .set_last_confirmed_frame(min_confirmed, self.sparse_saving);
 
-        // check time sync between clients and wait, if appropriate
+        // check time sync between clients and send wait recommendation, if appropriate
         if self.sync_layer.current_frame() > self.next_recommended_sleep {
             let skip_frames = self.max_delay_recommendation(true);
             if skip_frames > 0 {
@@ -328,8 +377,11 @@ impl P2PSession {
             }
         }
 
-        // save the current frame in the syncronization layer
-        requests.push(self.sync_layer.save_current_state());
+        // without sparse saving, always save the current frame
+        if !self.sparse_saving {
+            requests.push(self.sync_layer.save_current_state());
+        }
+
         // get correct inputs for the current frame
         let inputs = self
             .sync_layer
@@ -338,6 +390,7 @@ impl P2PSession {
             // check if input is correct or represents a disconnected player (by NULL_FRAME)
             assert!(input.frame == NULL_FRAME || input.frame == self.sync_layer.current_frame());
         }
+
         // advance the frame
         self.sync_layer.advance_frame();
         requests.push(GGRSRequest::AdvanceFrame { inputs });
@@ -345,10 +398,66 @@ impl P2PSession {
         Ok(requests)
     }
 
-    /// Used to fetch some statistics about the quality of the network connection.
+    /// Should be called periodically by your application to give GGRS a chance to do internal work.
+    /// GGRS will receive UDP packets, distribute them to corresponding endpoints, handle all occurring events and send all outgoing UDP packets.
+    pub fn poll_remote_clients(&mut self) {
+        // Get all udp packets and distribute them to associated endpoints.
+        // The endpoints will handle their packets, which will trigger both events and UPD replies.
+        for (from, msg) in &self.socket.receive_all_messages() {
+            for endpoint in self
+                .players
+                .values_mut()
+                .filter_map(Player::as_endpoint_mut)
+            {
+                if endpoint.is_handling_message(from) {
+                    endpoint.handle_message(msg);
+                    break;
+                }
+            }
+        }
+
+        // update frame information between remote players
+        for endpoint in self
+            .players
+            .values_mut()
+            .filter_map(Player::remote_as_endpoint_mut)
+        {
+            if endpoint.is_running() {
+                endpoint.update_local_frame_advantage(self.sync_layer.current_frame());
+            }
+        }
+
+        // run enpoint poll and get events from players and spectators. This will trigger additional UDP packets to be sent.
+        let mut events = VecDeque::new();
+        for endpoint in self
+            .players
+            .values_mut()
+            .filter_map(Player::as_endpoint_mut)
+        {
+            let player_handle = endpoint.player_handle();
+            for event in endpoint.poll(&self.local_connect_status) {
+                events.push_back((event, player_handle))
+            }
+        }
+
+        // handle all events locally
+        for (event, handle) in events.drain(..) {
+            self.handle_event(event, handle);
+        }
+
+        // send all queued UDP packets
+        for endpoint in self
+            .players
+            .values_mut()
+            .filter_map(Player::as_endpoint_mut)
+        {
+            endpoint.send_all_messages(&self.socket);
+        }
+    }
+
+    /// Returns a `NetworkStats` struct that gives information about the quality of the network connection.
     /// # Errors
-    /// - Returns `InvalidHandle` if the provided player handle is higher than the number of players.
-    /// - Returns `InvalidRequest` if the provided player handle does not refer to an existing remote player.
+    /// - Returns `InvalidHandle` if the provided player handle does not refer to an existing remote player.
     /// - Returns `NotSynchronized` if the session is not connected to other clients yet.
     pub fn network_stats(&self, player_handle: PlayerHandle) -> Result<NetworkStats, GGRSError> {
         // player handle is invalid
@@ -359,9 +468,11 @@ impl P2PSession {
         match self
             .players
             .get(&player_handle)
-            .ok_or(GGRSError::InvalidRequest)?
+            .ok_or(GGRSError::InvalidHandle)?
         {
-            Player::Local => Err(GGRSError::InvalidRequest),
+            Player::Local => Err(GGRSError::InvalidRequest {
+                info: "Cannot retrieve network statistics for the local player.".to_owned(),
+            }),
             Player::Remote(endpoint) | Player::Spectator(endpoint) => {
                 match endpoint.network_stats() {
                     Some(stats) => Ok(stats),
@@ -373,8 +484,8 @@ impl P2PSession {
 
     /// Change the amount of frames GGRS will delay the inputs for a player. You should only set the frame delay for local players.
     /// # Errors
-    /// Returns `InvalidHandle` if the provided player handle is higher than the number of players.
-    /// Returns `InvalidRequest` if the provided player handle refers to a remote player.
+    /// - Returns `InvalidHandle` if the provided player handle is invalid.
+    /// - Returns `InvalidRequest` if the provided player handle does not refer to a local player.
     pub fn set_frame_delay(
         &mut self,
         frame_delay: u32,
@@ -388,9 +499,11 @@ impl P2PSession {
         match self
             .players
             .get(&player_handle)
-            .ok_or(GGRSError::InvalidRequest)?
+            .ok_or(GGRSError::InvalidHandle)?
         {
-            Player::Remote(_) | Player::Spectator(_) => Err(GGRSError::InvalidRequest),
+            Player::Remote(_) | Player::Spectator(_) => Err(GGRSError::InvalidRequest {
+                info: "Frame delay can only be set for the local player.".to_owned(),
+            }),
             Player::Local => {
                 self.sync_layer.set_frame_delay(player_handle, frame_delay);
                 Ok(())
@@ -400,6 +513,7 @@ impl P2PSession {
 
     /// Sets the disconnect timeout. The session will automatically disconnect from a remote peer if it has not received a packet in the timeout window.
     pub fn set_disconnect_timeout(&mut self, timeout: Duration) {
+        self.disconnect_timeout = timeout;
         for endpoint in self
             .players
             .values_mut()
@@ -411,6 +525,7 @@ impl P2PSession {
 
     /// Sets the time before the first notification will be sent in case of a prolonged period of no received packages.
     pub fn set_disconnect_notify_delay(&mut self, notify_delay: Duration) {
+        self.disconnect_notify_start = notify_delay;
         for endpoint in self
             .players
             .values_mut()
@@ -420,6 +535,44 @@ impl P2PSession {
         }
     }
 
+    /// Sets the FPS this session is used with. This influences estimations for frame synchronization between sessions.
+    pub fn set_fps(&mut self, fps: u32) -> Result<(), GGRSError> {
+        if fps == 0 {
+            return Err(GGRSError::InvalidRequest {
+                info: "FPS should be higher than 0.".to_owned(),
+            });
+        }
+
+        self.fps = fps;
+
+        for endpoint in self
+            .players
+            .values_mut()
+            .filter_map(Player::as_endpoint_mut)
+        {
+            endpoint.set_fps(fps);
+        }
+
+        Ok(())
+    }
+
+    /// Sets the sparse saving mode. With sparse saving turned on, only the minimum confirmed frame (for which all inputs from all players are confirmed correct) will be saved.
+    /// This leads to much less save requests at the cost of potentially longer rollbacks and thus more advance frame requests. Recommended, if saving your gamestate
+    /// takes much more time than advancing the game state.
+    pub fn set_sparse_saving(&mut self, sparse_saving: bool) -> Result<(), GGRSError> {
+        // you can only switch the saving mode in the init phase
+        if self.state != SessionState::Initializing {
+            return Err(GGRSError::InvalidRequest {
+                info:
+                    "Session already started. You can only change the saving mode before starting the session."
+                        .to_owned(),
+            });
+        }
+
+        self.sparse_saving = sparse_saving;
+        Ok(())
+    }
+
     /// Returns the current `SessionState` of a session.
     pub const fn current_state(&self) -> SessionState {
         self.state
@@ -438,12 +591,14 @@ impl P2PSession {
 
         // check if player handle already exists
         if self.players.contains_key(&player_handle) {
-            return Err(GGRSError::InvalidRequest);
+            return Err(GGRSError::InvalidRequest {
+                info: "Player handle already exists.".to_owned(),
+            });
         }
 
         // check if a local player already exists
         if self.players.values().any(|p| matches!(p, Player::Local)) {
-            return Err(GGRSError::InvalidRequest);
+            return Err(GGRSError::InvalidRequest{info: "Local player already registered. It is not possible to add more than one local player.".to_owned()});
         }
 
         // finally add the local player
@@ -463,7 +618,9 @@ impl P2PSession {
 
         // check if player handle already exists
         if self.players.contains_key(&player_handle) {
-            return Err(GGRSError::InvalidRequest);
+            return Err(GGRSError::InvalidRequest {
+                info: "Player handle already exists.".to_owned(),
+            });
         }
 
         // create a udp protocol endpoint that handles all the messaging to that remote player
@@ -489,7 +646,9 @@ impl P2PSession {
 
         // check if player handle already exists
         if self.players.contains_key(&spectator_handle) {
-            return Err(GGRSError::InvalidRequest);
+            return Err(GGRSError::InvalidRequest {
+                info: "Player handle already exists.".to_owned(),
+            });
         }
 
         // create a udp protocol endpoint that handles all the messaging to that remote spectator
@@ -558,86 +717,56 @@ impl P2PSession {
         self.state = SessionState::Running;
     }
 
-    /// Receive UDP packages, distribute them to corresponding UDP endpoints, handle all occurring events and send all outgoing UDP packages.
-    /// Should be called periodically by your application to give GGRS a chance to do internal work like packet transmissions.
-    pub fn poll_remote_clients(&mut self) {
-        // Get all udp packets and distribute them to associated endpoints.
-        // The endpoints will handle their packets, which will trigger both events and UPD replies.
-        for (from, msg) in &self.socket.receive_all_messages() {
-            for endpoint in self
-                .players
-                .values_mut()
-                .filter_map(Player::as_endpoint_mut)
-            {
-                if endpoint.is_handling_message(from) {
-                    endpoint.handle_message(msg);
-                    break;
-                }
-            }
-        }
-
-        // update frame information between remote players
-        for endpoint in self
-            .players
-            .values_mut()
-            .filter_map(Player::remote_as_endpoint_mut)
-        {
-            if endpoint.is_running() {
-                endpoint.update_local_frame_advantage(self.sync_layer.current_frame());
-            }
-        }
-
-        // run enpoint poll and get events from players and spectators. This will trigger additional UDP packets to be sent.
-        let mut events = VecDeque::new();
-        for endpoint in self
-            .players
-            .values_mut()
-            .filter_map(Player::as_endpoint_mut)
-        {
-            let player_handle = endpoint.player_handle();
-            for event in endpoint.poll(&self.local_connect_status) {
-                events.push_back((event, player_handle))
-            }
-        }
-
-        // handle all events locally
-        for (event, handle) in events.drain(..) {
-            self.handle_event(event, handle);
-        }
-
-        // send all queued UDP packets
-        for endpoint in self
-            .players
-            .values_mut()
-            .filter_map(Player::as_endpoint_mut)
-        {
-            endpoint.send_all_messages(&self.socket);
-        }
-    }
-
-    /// Roll back to `first_incorrect` frame and resimulate the game with most up-to-date input data.
-    fn adjust_gamestate(&mut self, first_incorrect: Frame, requests: &mut Vec<GGRSRequest>) {
+    /// Roll back to `min_confirmed` frame and resimulate the game with most up-to-date input data.
+    fn adjust_gamestate(
+        &mut self,
+        first_incorrect: Frame,
+        min_confirmed: Frame,
+        requests: &mut Vec<GGRSRequest>,
+    ) {
         let current_frame = self.sync_layer.current_frame();
-        let count = current_frame - first_incorrect;
+        // determine the frame to load
+        let frame_to_load = if self.sparse_saving {
+            // if sparse saving is turned on, we will rollback to the last saved state
+            self.sync_layer.last_saved_frame()
+        } else {
+            // otherwise, we will rollback to first_incorrect
+            first_incorrect
+        };
 
-        // rollback to the first incorrect state
-        requests.push(self.sync_layer.load_frame(first_incorrect));
-        self.sync_layer.reset_prediction(first_incorrect);
-        assert_eq!(self.sync_layer.current_frame(), first_incorrect);
+        // we should always load a frame that is before or exactly the first incorrect frame
+        assert!(frame_to_load <= first_incorrect);
+        let count = current_frame - frame_to_load;
 
-        // step forward to the previous current state
-        for i in 0..count {
+        // request to load that frame
+        requests.push(self.sync_layer.load_frame(frame_to_load));
+
+        // we are now at the desired frame
+        assert_eq!(self.sync_layer.current_frame(), frame_to_load);
+        self.sync_layer.reset_prediction();
+
+        // step forward to the previous current state, but with updated inputs
+        for _ in 0..count {
             let inputs = self
                 .sync_layer
                 .synchronized_inputs(&self.local_connect_status);
 
-            if i > 0 {
-                requests.push(self.sync_layer.save_current_state());
-            }
-
+            // advance the frame
             self.sync_layer.advance_frame();
             requests.push(GGRSRequest::AdvanceFrame { inputs });
+
+            // decide wether to request a state save
+            if self.sparse_saving {
+                // with sparse saving, we only save exactly the min_confirmed frame
+                if self.sync_layer.current_frame() == min_confirmed {
+                    requests.push(self.sync_layer.save_current_state());
+                }
+            } else {
+                // without sparse saving, we save every state except the very first one
+                requests.push(self.sync_layer.save_current_state());
+            }
         }
+        // after all this, we should have arrived at the same frame where we started
         assert_eq!(self.sync_layer.current_frame(), current_frame);
     }
 
@@ -691,11 +820,7 @@ impl P2PSession {
             let mut queue_min_confirmed = i32::MAX;
 
             // check all remote players for that player
-            for endpoint in self
-                .players
-                .values_mut()
-                .filter_map(Player::remote_as_endpoint_mut)
-            {
+            for endpoint in self.players.values().filter_map(Player::remote_as_endpoint) {
                 if !endpoint.is_running() {
                     continue;
                 }
