@@ -4,8 +4,7 @@ use crate::error::GGRSError;
 use crate::frame_info::GameInput;
 use crate::network::udp_msg::ConnectionStatus;
 use crate::sync_layer::SyncLayer;
-use crate::{Frame, GGRSRequest};
-use crate::{PlayerHandle, PlayerType, SessionState};
+use crate::{Frame, GGRSRequest, PlayerHandle};
 
 /// During a `SyncTestSession`, GGRS will simulate a rollback every frame and resimulate the last n states, where n is the given check distance.
 /// The resimulated checksums will be compared with the original checksums and report if there was a mismatch.
@@ -14,7 +13,6 @@ pub struct SyncTestSession {
     num_players: u32,
     input_size: usize,
     check_distance: u32,
-    running: bool,
     sync_layer: SyncLayer,
     dummy_connect_status: Vec<ConnectionStatus>,
     checksum_history: HashMap<Frame, u64>,
@@ -31,72 +29,27 @@ impl SyncTestSession {
             num_players,
             input_size,
             check_distance,
-            running: false,
             sync_layer: SyncLayer::new(num_players, input_size),
             dummy_connect_status,
             checksum_history: HashMap::default(),
         }
     }
 
-    /// Must be called for each player in the session (e.g. in a 3 player session, must be called 3 times).
-    /// # Errors
-    /// Will return `InvalidHandle` when the provided player handle is too big for the number of players.
-    /// Will return `InvalidRequest` if a player with that handle has been added before.
-    /// Will return `InvalidRequest` for any player type other than `Local`. `SyncTestSession` does not support remote players.
-    pub fn add_player(
-        &mut self,
-        player_type: PlayerType,
-        player_handle: PlayerHandle,
-    ) -> Result<(), GGRSError> {
-        if player_handle >= self.num_players as PlayerHandle {
-            return Err(GGRSError::InvalidHandle);
-        }
-        if player_type != PlayerType::Local {
-            return Err(GGRSError::InvalidRequest);
-        }
-        Ok(())
-    }
-
-    /// After you are done defining and adding all players, you should start the session. In a sync test, starting the session saves the initial game state and sets running to true.
-    ///
-    /// # Errors
-    /// Return a `InvalidRequestError`, if the session is already running.
-    pub fn start_session(&mut self) -> Result<(), GGRSError> {
-        if self.running {
-            return Err(GGRSError::InvalidRequest);
-        }
-
-        self.running = true;
-        Ok(())
-    }
-
     /// In a sync test, this will advance the state by a single frame and afterwards rollback `check_distance` amount of frames,
     /// resimulate and compare checksums with the original states. Returns an order-sensitive `Vec<GGRSRequest>`.
     /// You should fulfill all requests in the exact order they are provided. Failure to do so will cause panics later.
     ///
     /// # Errors
-    /// - Returns `InvalidHandle` if the provided player handle is higher than the number of players.
     /// - Returns `MismatchedChecksumError` if checksums don't match after resimulation.
-    /// - Returns `NotSynchronized` if the session has not been started yet.
     pub fn advance_frame(
         &mut self,
-        player_handle: PlayerHandle,
-        input: &[u8],
+        all_inputs: &Vec<Vec<u8>>,
     ) -> Result<Vec<GGRSRequest>, GGRSError> {
-        // player handle is invalid
-        if player_handle > self.num_players as PlayerHandle {
-            return Err(GGRSError::InvalidHandle);
-        }
-        // session has not been started
-        if !self.running {
-            return Err(GGRSError::NotSynchronized);
-        }
-
         let mut requests = Vec::new();
 
         // if we advanced far enough into the game do comparisons and rollbacks
         if self.check_distance > 0 && self.sync_layer.current_frame() > self.check_distance as i32 {
-            // compare checksums of older frames to our checksum history (where only the first version of any checksum is allowed)
+            // compare checksums of older frames to our checksum history (where only the first version of any checksum is recorded)
             for i in 0..=self.check_distance as i32 {
                 let frame_to_check = self.sync_layer.current_frame() - i;
                 if !self.checksums_consistent(frame_to_check) {
@@ -111,14 +64,17 @@ impl SyncTestSession {
             self.adjust_gamestate(frame_to, &mut requests);
         }
 
-        //create an input struct for current frame
-        let mut current_input: GameInput =
-            GameInput::new(self.sync_layer.current_frame(), self.input_size);
-        current_input.copy_input(input);
+        // pass all inputs into the sync layer
+        assert_eq!(self.num_players as usize, all_inputs.len());
+        for i in 0..self.num_players as usize {
+            //create an input struct for current frame
+            let mut input: GameInput =
+                GameInput::new(self.sync_layer.current_frame(), self.input_size);
+            input.copy_input(&all_inputs[i]);
 
-        // send the input into the sync layer
-        self.sync_layer
-            .add_local_input(player_handle, current_input)?;
+            // send the input into the sync layer
+            self.sync_layer.add_local_input(i, input)?;
+        }
 
         // save the current frame in the syncronization layer
         requests.push(self.sync_layer.save_current_state());
@@ -139,7 +95,7 @@ impl SyncTestSession {
         // inputs from other players
         let safe_frame = self.sync_layer.current_frame() - self.check_distance as i32;
 
-        self.sync_layer.set_last_confirmed_frame(safe_frame);
+        self.sync_layer.set_last_confirmed_frame(safe_frame, false);
 
         // also, we update the dummy connect status to pretend that we received inputs from all players
         for con_stat in &mut self.dummy_connect_status {
@@ -166,15 +122,6 @@ impl SyncTestSession {
         Ok(())
     }
 
-    /// Returns the current state of the `SynctestSession`.
-    pub const fn current_state(&self) -> SessionState {
-        if self.running {
-            SessionState::Running
-        } else {
-            SessionState::Initializing
-        }
-    }
-
     /// Updates the `checksum_history` and checks if the checksum is identical if it already has been recorded once
     fn checksums_consistent(&mut self, frame_to_check: Frame) -> bool {
         // remove entries older than the check_distance
@@ -200,27 +147,23 @@ impl SyncTestSession {
     }
 
     fn adjust_gamestate(&mut self, frame_to: Frame, requests: &mut Vec<GGRSRequest>) {
-        let current_frame = self.sync_layer.current_frame();
-        let count = current_frame - frame_to;
+        let start_frame = self.sync_layer.current_frame();
+        let count = start_frame - frame_to;
 
         // rollback to the first incorrect state
         requests.push(self.sync_layer.load_frame(frame_to));
-        self.sync_layer.reset_prediction(frame_to);
+        self.sync_layer.reset_prediction();
         assert_eq!(self.sync_layer.current_frame(), frame_to);
 
         // step forward to the previous current state
-        for i in 0..count {
+        for _ in 0..count {
             let inputs = self
                 .sync_layer
                 .synchronized_inputs(&self.dummy_connect_status);
 
-            if i > 0 {
-                requests.push(self.sync_layer.save_current_state());
-            }
-
             self.sync_layer.advance_frame();
             requests.push(GGRSRequest::AdvanceFrame { inputs });
         }
-        assert_eq!(self.sync_layer.current_frame(), current_frame);
+        assert_eq!(self.sync_layer.current_frame(), start_frame);
     }
 }
