@@ -59,13 +59,11 @@ pub(crate) struct SavedStates {
     // the states array is two bigger than the max prediction frames in order to account for
     // the next frame needing a space and still being able to rollback the max distance
     pub states: [GameStateCell; MAX_PREDICTION_FRAMES as usize + 2],
-    pub head: usize,
 }
 
 impl Default for SavedStates {
     fn default() -> Self {
         Self {
-            head: 0,
             states: Default::default(),
         }
     }
@@ -73,26 +71,18 @@ impl Default for SavedStates {
 
 impl SavedStates {
     fn push(&mut self, frame: Frame) -> GameStateCell {
-        let saved_state = self.states[self.head].clone();
-        saved_state.reset(frame);
-        self.head = (self.head + 1) % self.states.len();
-        assert!(self.head < self.states.len());
-        saved_state
+        assert!(frame >= 0);
+        let pos = frame as usize % self.states.len();
+        let cell = self.states[pos].clone();
+        cell.reset(frame);
+        cell
     }
 
-    fn find_index(&self, frame: Frame) -> Option<usize> {
-        self.states
-            .iter()
-            .enumerate()
-            .find(|(_, saved)| saved.0.lock().frame == frame)
-            .map(|(i, _)| i)
-    }
-
-    fn reset_to(&mut self, frame: Frame) -> GameStateCell {
-        self.head = self
-            .find_index(frame)
-            .unwrap_or_else(|| panic!("Could not find saved frame index for frame: {}", frame));
-        self.states[self.head].clone()
+    fn peek(&mut self, frame: Frame) -> GameStateCell {
+        assert!(frame >= 0);
+        let pos = frame as usize % self.states.len();
+        let saved_cell = self.states[pos].clone();
+        saved_cell
     }
 
     fn by_frame(&self, frame: Frame) -> Option<GameStateCell> {
@@ -110,6 +100,7 @@ pub(crate) struct SyncLayer {
     saved_states: SavedStates,
     rolling_back: bool,
     last_confirmed_frame: Frame,
+    last_saved_frame: Frame,
     current_frame: Frame,
     input_queues: Vec<InputQueue>,
 }
@@ -127,9 +118,9 @@ impl SyncLayer {
             input_size,
             rolling_back: false,
             last_confirmed_frame: NULL_FRAME,
+            last_saved_frame: NULL_FRAME,
             current_frame: 0,
             saved_states: SavedStates {
-                head: 0,
                 states: Default::default(),
             },
             input_queues,
@@ -145,6 +136,7 @@ impl SyncLayer {
     }
 
     pub(crate) fn save_current_state(&mut self) -> GGRSRequest {
+        self.last_saved_frame = self.current_frame;
         let cell = self.saved_states.push(self.current_frame);
         GGRSRequest::SaveGameState {
             cell,
@@ -157,13 +149,13 @@ impl SyncLayer {
         self.input_queues[player_handle as usize].set_frame_delay(delay);
     }
 
-    pub(crate) fn reset_prediction(&mut self, frame: Frame) {
+    pub(crate) fn reset_prediction(&mut self) {
         for i in 0..self.num_players {
-            self.input_queues[i as usize].reset_prediction(frame);
+            self.input_queues[i as usize].reset_prediction();
         }
     }
 
-    /// Loads the gamestate indicated by `frame_to_load`. After execution, `self.saved_states.head` is set one position after the loaded state.
+    /// Loads the gamestate indicated by `frame_to_load`.
     pub(crate) fn load_frame(&mut self, frame_to_load: Frame) -> GGRSRequest {
         // The state should not be the current state or the state should not be in the future or too far away in the past
         assert!(
@@ -173,11 +165,10 @@ impl SyncLayer {
         );
 
         // Reset the head of the state ring-buffer to point in advance of the current frame (as if we had just finished executing it).
-        let cell = self.saved_states.reset_to(frame_to_load);
+        let cell = self.saved_states.peek(frame_to_load);
         let loaded_frame = cell.0.lock().frame;
         assert_eq!(loaded_frame, frame_to_load);
 
-        self.saved_states.head = (self.saved_states.head + 1) % MAX_PREDICTION_FRAMES as usize;
         self.current_frame = loaded_frame;
 
         GGRSRequest::LoadGameState { cell }
@@ -213,7 +204,7 @@ impl SyncLayer {
     ) -> Vec<GameInput> {
         let mut inputs = Vec::new();
         for (i, con_stat) in connect_status.iter().enumerate() {
-            if con_stat.disconnected {
+            if con_stat.disconnected && con_stat.last_frame < self.current_frame {
                 inputs.push(BLANK_INPUT);
             } else {
                 inputs.push(self.input_queues[i].input(self.current_frame));
@@ -230,17 +221,17 @@ impl SyncLayer {
     ) -> Vec<GameInput> {
         let mut inputs = Vec::new();
         for (i, con_stat) in connect_status.iter().enumerate() {
-            if con_stat.disconnected || con_stat.last_frame < frame {
+            if con_stat.disconnected && con_stat.last_frame < frame {
                 inputs.push(BLANK_INPUT);
             } else {
-                inputs.push(self.input_queues[i].confirmed_input(frame as u32));
+                inputs.push(self.input_queues[i].confirmed_input(frame));
             }
         }
         inputs
     }
 
     /// Sets the last confirmed frame to a given frame. By raising the last confirmed frame, we can discard all previous frames, as they are no longer necessary.
-    pub(crate) fn set_last_confirmed_frame(&mut self, frame: Frame) {
+    pub(crate) fn set_last_confirmed_frame(&mut self, mut frame: Frame, sparse_saving: bool) {
         // dont set the last confirmed frame after the first incorrect frame before a rollback has happened
         let mut first_incorrect: Frame = NULL_FRAME;
         for handle in 0..self.num_players as usize {
@@ -250,6 +241,11 @@ impl SyncLayer {
             );
         }
 
+        // if sparse saving option is turned on, don't set the last confirmed frame after the last saved frame
+        if sparse_saving {
+            frame = std::cmp::min(frame, self.last_saved_frame);
+        }
+
         // if we set the last confirmed frame beyond the first incorrect frame, we discard inputs that we need later for ajusting the gamestate.
         assert!(first_incorrect == NULL_FRAME || first_incorrect >= frame);
 
@@ -262,7 +258,7 @@ impl SyncLayer {
     }
 
     /// Finds the earliest incorrect frame detected by the individual input queues
-    pub(crate) fn check_simulation_consistency(&self, mut first_incorrect: Frame) -> Option<Frame> {
+    pub(crate) fn check_simulation_consistency(&self, mut first_incorrect: Frame) -> Frame {
         for handle in 0..self.num_players as usize {
             let incorrect = self.input_queues[handle].first_incorrect_frame();
             if incorrect != NULL_FRAME
@@ -271,16 +267,18 @@ impl SyncLayer {
                 first_incorrect = incorrect;
             }
         }
-        match first_incorrect {
-            NULL_FRAME => None,
-            _ => Some(first_incorrect),
-        }
+        first_incorrect
     }
 
-    /// Returns the latest saved gamestate
+    /// Returns a gamestate through given frame
     pub(crate) fn saved_state_by_frame(&self, frame: Frame) -> Option<GameStateCell> {
         self.saved_states.by_frame(frame)
     }
+
+    /// Returns the latest saved frame
+    pub(crate) const fn last_saved_frame(&self) -> Frame {
+        self.last_saved_frame
+    }
 }
 
 // #########
